#include <cassert>
#include <cstring>
#include <cstdio>
#include <map>
#include <list>
#include <vector>
#include <algorithm>
#include <functional>
#include <iostream>
#include <FreeImagePlus.h>
#include "label.h"
#include "scalarutils.h"

class DisjointSet {
public:
    struct Set {
        unsigned int parent;
        unsigned int rank;
    };

    DisjointSet(unsigned int initialSets)
        : initialSets(initialSets), sets(new Set[initialSets]) {
        for (unsigned int i = 0; i < initialSets; ++i) {
            this->sets[i].parent = i;
            this->sets[i].rank   = 0;
        }
    }

    virtual ~DisjointSet() {
        delete [] this->sets;
    }

    unsigned int findRepresentativeSet(unsigned int x) {
        if (x != this->sets[x].parent) {
            this->sets[x].parent = this->findRepresentativeSet(this->sets[x].parent);
        }
        return this->sets[x].parent;
    }

    void mergeSets(unsigned int x, unsigned int y) {
        unsigned int xRoot = this->findRepresentativeSet(x);
        unsigned int yRoot = this->findRepresentativeSet(y);
        if (xRoot != yRoot) {
            // x and y are not already in same set. Merge them.
            if (this->sets[xRoot].rank < this->sets[yRoot].rank) {
                this->sets[xRoot].parent = yRoot;
            } else {
                this->sets[yRoot].parent = xRoot;
                if (this->sets[xRoot].rank == this->sets[yRoot].rank) {
                    (this->sets[xRoot].rank)++;
                }
            }
        }
    }

private:
    DisjointSet();
    unsigned int initialSets;
    Set * const sets;
};








//template <typename T>
//unsigned int labelImage(RegionInfo    regionInfo[numMaxRegions + 1],
//                unsigned int * labelled,
//                const T      * probBG,
//                unsigned int   imageHeight,
//                unsigned int   imageWidth) {
//    DisjointSet equivalencies((imageHeight / 2 + 1) * (imageWidth / 2 + 1));
//    unsigned int newLabel = 1;
//    for (int r = 0; r < static_cast<int>(imageHeight); ++r) {
//        for (int c = 0; c < static_cast<int>(imageWidth); ++c) {
//            unsigned int foundLabel = 0;
//            if (*displacePtr3D (probBG, imageWidth, imageHeight, c, r, 0)) {
//                unsigned int nwLabel = r - 1 >= 0 && c - 1 >= 0                            && *displacePtr3D (probBG, imageWidth, imageHeight, c - 1, r - 1, 0) < (ScalarInfo<T>::NORM_MAX / 2) ? *displacePtr3D (labelled, imageWidth, imageHeight, c - 1, r - 1, 0) : 0;
//                unsigned int nLabel  = r - 1 >= 0 &&                                          *displacePtr3D (probBG, imageWidth, imageHeight, c    , r - 1, 0) < (ScalarInfo<T>::NORM_MAX / 2) ? *displacePtr3D (labelled, imageWidth, imageHeight, c    , r - 1, 0) : 0;
//                unsigned int neLabel = r - 1 >= 0 && c + 1  < static_cast<int>(imageWidth) && *displacePtr3D (probBG, imageWidth, imageHeight, c + 1, r - 1, 0) < (ScalarInfo<T>::NORM_MAX / 2) ? *displacePtr3D (labelled, imageWidth, imageHeight, c + 1, r - 1, 0) : 0;
//                unsigned int wLabel  =               c - 1 >= 0                            && *displacePtr3D (probBG, imageWidth, imageHeight, c - 1, r    , 0) < (ScalarInfo<T>::NORM_MAX / 2) ? *displacePtr3D (labelled, imageWidth, imageHeight, c - 1, r    , 0) : 0;
//                unsigned int neighbourLabels[4] = {nwLabel, nLabel, neLabel, wLabel};
//                for (int l = 0; l < 4; ++l) {
//                    if (0 < neighbourLabels[l]) {
//                        if (0 == foundLabel) {
//                            foundLabel = neighbourLabels[l];
//                        } else {
//                            equivalencies.mergeSets(foundLabel, neighbourLabels[l]);
//                        }
//                    }
//                }
//                if (0 == foundLabel) {
//                    foundLabel = newLabel;
//                    ++newLabel;
//                }
//                *displacePtr3D (labelled, imageWidth, imageHeight, c, r, 0) = equivalencies.findRepresentativeSet(foundLabel);
//            } else {
//                *displacePtr3D (labelled, imageWidth, imageHeight, c, r, 0) = 0;
//            }
//        }
//    }

//    typedef std::map<unsigned int, unsigned int> Uint2UintMap;
//    Uint2UintMap labelSizeMap;
//    for (int r = 0; r < static_cast<int>(imageHeight); ++r) {
//        for (int c = 0; c < static_cast<int>(imageWidth); ++c) {
//            if (0 < *displacePtr3D (labelled, imageWidth, imageHeight, c, r, 0)) {
//                unsigned int representativeLabel = equivalencies.findRepresentativeSet(*displacePtr3D (labelled, imageWidth, imageHeight, c, r, 0));
//                *displacePtr3D (labelled, imageWidth, imageHeight, c, r, 0) = representativeLabel;
//                labelSizeMap[representativeLabel] = labelSizeMap[representativeLabel] + 1;
//            }
//        }
//    }


//    //unsigned int numRegions = labelSizeMap.size();
//    typedef std::pair<unsigned int, unsigned int> SizeLabelPair;
//    typedef std::list<SizeLabelPair> SizeLabelPairList;
//    SizeLabelPairList sizeLabelPairList;
//    for (Uint2UintMap::iterator it = labelSizeMap.begin(); it != labelSizeMap.end(); ++it) {
//        //printf("Label %d: %u occurrences\n", it->first, it->second);
//        sizeLabelPairList.push_back(SizeLabelPair(it->second, it->first));
//    }
//    sizeLabelPairList.sort(std::greater<SizeLabelPair>());
//    //unsigned int * pixelsPerRegion = new unsigned int[numRegions + 1];
//    //memset(pixelsPerRegion, 0, sizeof(unsigned int) * (numRegions + 1));


//    unsigned int conformingRegions = 0;

//    regionInfo[0].label = 0;
//    regionInfo[0].numPixels = 0;
////    regionInfo[0].covered = false;

//    unsigned int numberOfPixelsInConformingRegions = 0;

//    Uint2UintMap newLabels;
//    {   unsigned int i;
//        SizeLabelPairList::iterator it;
//        for (i = 1, it = sizeLabelPairList.begin();
//             it != sizeLabelPairList.end();
//             ++it, ++i) {
//            if ((i <= numMaxRegions) && (minRegionSize <= it->first)) {
//                regionInfo[i].label = i;
//                regionInfo[i].numPixels = it->first;
//                newLabels[it->second] = i;
//                conformingRegions++;
////                regionInfo[i].covered = false; // This will be determined later on
//                numberOfPixelsInConformingRegions += regionInfo[i].numPixels;
//            } else {
//                if (i <= numMaxRegions) {
//                    regionInfo[i].label = 0;
//                    regionInfo[i].numPixels = 0;
////                    regionInfo[i].covered = false; // Irrelevant since it's a non-valid region
//                }
//                newLabels[it->second] = 0;
//            }
//        }
//    }

//    std::vector<int> rowPositions[numMaxRegions + 1];
//    std::vector<int> colPositions[numMaxRegions + 1];

//    for (unsigned int i = 1; i <= conformingRegions; ++i) {
//        regionInfo[i].maxRow = 0;
//        regionInfo[i].minRow = UINT_MAX;
//        regionInfo[i].maxCol = 0;
//        regionInfo[i].minCol = UINT_MAX;
//    }

//    std::vector<PixelInRegion> pixelsInConformingRegions;
//    pixelsInConformingRegions.reserve(numberOfPixelsInConformingRegions);

//    for (unsigned int r = 0; r < imageHeight; ++r) {
//        for (unsigned int c = 0; c < imageWidth; ++c) {
//            if (0 < *displacePtr3D (labelled, imageWidth, imageHeight, c, r, 0)) {
//                const unsigned int oldLabel = equivalencies.findRepresentativeSet(*displacePtr3D (labelled, imageWidth, imageHeight, c, r, 0));
//                const unsigned int definitiveLabel = newLabels[oldLabel];
//                *displacePtr3D (labelled, imageWidth, imageHeight, c, r, 0) = definitiveLabel;
//                PixelInRegion px;
//                px.row = r;
//                px.col = c;
//                px.label = definitiveLabel;
//                pixelsInConformingRegions.push_back(px);

//                rowPositions[definitiveLabel].push_back(r);
//                colPositions[definitiveLabel].push_back(c);
//                regionInfo[definitiveLabel].maxRow = std::max(regionInfo[definitiveLabel].maxRow, r);
//                regionInfo[definitiveLabel].minRow = std::min(regionInfo[definitiveLabel].minRow, r);
//                regionInfo[definitiveLabel].maxCol = std::max(regionInfo[definitiveLabel].maxCol, c);
//                regionInfo[definitiveLabel].minCol = std::min(regionInfo[definitiveLabel].minCol, c);
//            }
//        }
//    }

//    for (unsigned int i = 1; i <= conformingRegions; ++i) {
//        std::nth_element(rowPositions[i].begin(), rowPositions[i].begin() + rowPositions[i].size()/2, rowPositions[i].end());
////        std::cout << "The median (row) is " << (rowPositions[i])[rowPositions[i].size()/2] << '\n';
//        regionInfo[i].medianRow = (rowPositions[i])[rowPositions[i].size()/2];
//        std::nth_element(colPositions[i].begin(), colPositions[i].begin() + colPositions[i].size()/2, colPositions[i].end());
////        std::cout << "The median (col) is " << (colPositions[i])[colPositions[i].size()/2] << '\n';
//        regionInfo[i].medianCol = (colPositions[i])[colPositions[i].size()/2];
//    }

////    for (unsigned int i = 0; i < conformingRegions; ++i) {
////        printf("Label %d: %u occurrences\n", i+1, regionInfo[i + 1].numPixels);
////    }
//    return conformingRegions;
//}


/// Takes the grayscale image probBG, binarizes it with threshold: true >= 0.5
/// and detects connected components, then sorts them by size, assigning the
/// label 1 to the largest component.
///
/// It disregards regions with less than minRegionSize (global const) pixels
/// and will only consider the largest numMaxRegions (global const) in the image.
///
/// Upon returning, regionInfo will feature in its positions 1 to n the information
/// about the n valid regions found: area, bounding box, centre (median in each dim).
/// pixelsInConformingRegions will be filled with a list of all pixels belonging
/// to any of the found regions, in no particular order.
/// Finally, the function call returns the number of valid regions that were found.
template <typename T>
unsigned int linearLabelImage(
                unsigned int numMaxRegions,
                RegionInfo      regionInfo[],//[numMaxRegions + 1],
                unsigned int   minRegionSize,
                std::vector<PixelInRegion>  & pixelsInConformingRegions,
                const T      * probBG,
                unsigned int   imageHeight,
                unsigned int   imageWidth) {
    DisjointSet equivalencies((imageHeight / 2 + 1) * (imageWidth / 2 + 1));

    unsigned int * labelled = new unsigned int[imageHeight * imageWidth];

    unsigned int newLabel = 1;
    for (int r = 0; r < static_cast<int>(imageHeight); ++r) {
        for (int c = 0; c < static_cast<int>(imageWidth); ++c) {
            unsigned int foundLabel = 0;
            if (*displacePtr3D (probBG, imageWidth, imageHeight, c, r, 0)) {
                unsigned int nwLabel = r - 1 >= 0 && c - 1 >= 0                            && *displacePtr3D (probBG, imageWidth, imageHeight, c - 1, r - 1, 0) < (ScalarInfo<T>::NORM_MAX / 2) ? *displacePtr3D (labelled, imageWidth, imageHeight, c - 1, r - 1, 0) : 0;
                unsigned int nLabel  = r - 1 >= 0 &&                                          *displacePtr3D (probBG, imageWidth, imageHeight, c    , r - 1, 0) < (ScalarInfo<T>::NORM_MAX / 2) ? *displacePtr3D (labelled, imageWidth, imageHeight, c    , r - 1, 0) : 0;
                unsigned int neLabel = r - 1 >= 0 && c + 1  < static_cast<int>(imageWidth) && *displacePtr3D (probBG, imageWidth, imageHeight, c + 1, r - 1, 0) < (ScalarInfo<T>::NORM_MAX / 2) ? *displacePtr3D (labelled, imageWidth, imageHeight, c + 1, r - 1, 0) : 0;
                unsigned int wLabel  =               c - 1 >= 0                            && *displacePtr3D (probBG, imageWidth, imageHeight, c - 1, r    , 0) < (ScalarInfo<T>::NORM_MAX / 2) ? *displacePtr3D (labelled, imageWidth, imageHeight, c - 1, r    , 0) : 0;
                unsigned int neighbourLabels[4] = {nwLabel, nLabel, neLabel, wLabel};
                for (int l = 0; l < 4; ++l) {
                    if (0 < neighbourLabels[l]) {
                        if (0 == foundLabel) {
                            foundLabel = neighbourLabels[l];
                        } else {
                            equivalencies.mergeSets(foundLabel, neighbourLabels[l]);
                        }
                    }
                }
                if (0 == foundLabel) {
                    foundLabel = newLabel;
                    ++newLabel;
                }
                *displacePtr3D (labelled, imageWidth, imageHeight, c, r, 0) = equivalencies.findRepresentativeSet(foundLabel);
            } else {
                *displacePtr3D (labelled, imageWidth, imageHeight, c, r, 0) = 0;
            }
        }
    }

    typedef std::map<unsigned int, unsigned int> Uint2UintMap;
    Uint2UintMap labelSizeMap;
    for (int r = 0; r < static_cast<int>(imageHeight); ++r) {
        for (int c = 0; c < static_cast<int>(imageWidth); ++c) {
            if (0 < *displacePtr3D (labelled, imageWidth, imageHeight, c, r, 0)) {
                unsigned int representativeLabel = equivalencies.findRepresentativeSet(*displacePtr3D (labelled, imageWidth, imageHeight, c, r, 0));
                *displacePtr3D (labelled, imageWidth, imageHeight, c, r, 0) = representativeLabel;
                labelSizeMap[representativeLabel] = labelSizeMap[representativeLabel] + 1;
            }
        }
    }


    //unsigned int numRegions = labelSizeMap.size();
    typedef std::pair<unsigned int, unsigned int> SizeLabelPair;
    typedef std::list<SizeLabelPair> SizeLabelPairList;
    SizeLabelPairList sizeLabelPairList;
    for (Uint2UintMap::iterator it = labelSizeMap.begin(); it != labelSizeMap.end(); ++it) {
        //printf("Label %d: %u occurrences\n", it->first, it->second);
        sizeLabelPairList.push_back(SizeLabelPair(it->second, it->first));
    }
    sizeLabelPairList.sort(std::greater<SizeLabelPair>());
    //unsigned int * pixelsPerRegion = new unsigned int[numRegions + 1];
    //memset(pixelsPerRegion, 0, sizeof(unsigned int) * (numRegions + 1));


    unsigned int conformingRegions = 0;

    regionInfo[0].label = 0;
    regionInfo[0].numPixels = 0;
//    regionInfo[0].covered = false;

    unsigned int numberOfPixelsInConformingRegions = 0;

    Uint2UintMap newLabels;
    {   unsigned int i;
        SizeLabelPairList::iterator it;
        for (i = 1, it = sizeLabelPairList.begin();
             it != sizeLabelPairList.end();
             ++it, ++i) {
            if ((i <= numMaxRegions) && (minRegionSize <= it->first)) {
                regionInfo[i].label = i;
                regionInfo[i].numPixels = it->first;
                newLabels[it->second] = i;
                conformingRegions++;
//                regionInfo[i].covered = false; // This will be determined later on
                numberOfPixelsInConformingRegions += regionInfo[i].numPixels;
            } else {
                if (i <= numMaxRegions) {
                    regionInfo[i].label = 0;
                    regionInfo[i].numPixels = 0;
//                    regionInfo[i].covered = false; // Irrelevant since it's a non-valid region
                }
                newLabels[it->second] = 0;
            }
        }
    }

    std::vector<int> rowPositions[numMaxRegions + 1];
    std::vector<int> colPositions[numMaxRegions + 1];

    for (unsigned int i = 1; i <= conformingRegions; ++i) {
        regionInfo[i].maxRow = 0;
        regionInfo[i].minRow = UINT_MAX;
        regionInfo[i].maxCol = 0;
        regionInfo[i].minCol = UINT_MAX;
    }

    //std::vector<PixelInRegion> pixelsInConformingRegions;
    pixelsInConformingRegions.clear();
    pixelsInConformingRegions.reserve(numberOfPixelsInConformingRegions);

    for (unsigned int r = 0; r < imageHeight; ++r) {
        for (unsigned int c = 0; c < imageWidth; ++c) {
            if (0 < *displacePtr3D (labelled, imageWidth, imageHeight, c, r, 0)) {
                const unsigned int oldLabel = equivalencies.findRepresentativeSet(*displacePtr3D (labelled, imageWidth, imageHeight, c, r, 0));
                const unsigned int definitiveLabel = newLabels[oldLabel];
                *displacePtr3D (labelled, imageWidth, imageHeight, c, r, 0) = definitiveLabel;
                /// Since we have filtered out too many or too small regions, not every pixel we
                /// originally labelled will end up having a label
                if (0 < definitiveLabel) {
                    PixelInRegion px;
                    px.row = r;
                    px.col = c;
                    px.label = definitiveLabel;
                    //px.closestStateVector = 0; // No state vector is assigned yet
                    pixelsInConformingRegions.push_back(px);
                    rowPositions[definitiveLabel].push_back(r);
                    colPositions[definitiveLabel].push_back(c);
                    regionInfo[definitiveLabel].maxRow = std::max(regionInfo[definitiveLabel].maxRow, r);
                    regionInfo[definitiveLabel].minRow = std::min(regionInfo[definitiveLabel].minRow, r);
                    regionInfo[definitiveLabel].maxCol = std::max(regionInfo[definitiveLabel].maxCol, c);
                    regionInfo[definitiveLabel].minCol = std::min(regionInfo[definitiveLabel].minCol, c);
                }
            }
        }
    }

    delete [] labelled;

    for (unsigned int i = 1; i <= conformingRegions; ++i) {
        std::nth_element(rowPositions[i].begin(), rowPositions[i].begin() + rowPositions[i].size()/2, rowPositions[i].end());
//        std::cout << "The median (row) is " << (rowPositions[i])[rowPositions[i].size()/2] << '\n';
        regionInfo[i].medianRow = (rowPositions[i])[rowPositions[i].size()/2];
        std::nth_element(colPositions[i].begin(), colPositions[i].begin() + colPositions[i].size()/2, colPositions[i].end());
//        std::cout << "The median (col) is " << (colPositions[i])[colPositions[i].size()/2] << '\n';
        regionInfo[i].medianCol = (colPositions[i])[colPositions[i].size()/2];
    }

//    for (unsigned int i = 0; i < conformingRegions; ++i) {
//        printf("Label %d: %u occurrences\n", i+1, regionInfo[i + 1].numPixels);
//    }
    return conformingRegions;
}



//template
//unsigned int labelImage<unsigned char>(RegionInfo    regionInfo[numMaxRegions + 1],
//                               unsigned int        * annotated,
//                               const unsigned char * binary,
//                               unsigned int          imageHeight,
//                               unsigned int          imageWidth);

template
unsigned int linearLabelImage<unsigned char>(
unsigned int numMaxRegions,
RegionInfo      regionInfo[],//[numMaxRegions + 1],
unsigned int   minRegionSize,
std::vector<PixelInRegion>  & pixelsInConformingRegions,
const unsigned char      * probBG,
unsigned int   imageHeight,
unsigned int   imageWidth);
