#ifndef PROCESSINGDATATYPES_H
#define PROCESSINGDATATYPES_H

#include <cstddef>
#include <cassert>
#include <cuda_runtime.h>
#include "auxcuda.h"
#include "thumbnail.h"
#include "particlefilter.h"
#include "label.h"

template <typename T, unsigned int CHANNELS>
class ForegroundReferenceData;

struct ProcessingDeviceAux {
    float * backgroundSpatialHorizontalGaussian;
    float * backgroundSpatialVerticalGaussian;
    float * foregroundSpatialHorizontalGaussian;
    float * foregroundSpatialVerticalGaussian;

    ProcessingDeviceAux ()
        : backgroundSpatialHorizontalGaussian(NULL),
          backgroundSpatialVerticalGaussian(NULL),
          foregroundSpatialHorizontalGaussian(NULL),
          foregroundSpatialVerticalGaussian(NULL)
    {}

    virtual ~ProcessingDeviceAux ();
private:
    ProcessingDeviceAux (const ProcessingDeviceAux &other);
};

template <typename T>
struct ProcessingDeviceData {
//    T             * imageAbsDiffs;
    unsigned char * bgProbabilities;
//    long          * imageSquaredDiffTemporalWeightedSum;
//    long          * probTemporalSum;
    unsigned char * binsHistory; // This stores which bins were incremented in a particular moment so that they can be decremented when
    unsigned short * diffsHistograms; // One histogram per channel
    float          * medianSigmas; // This stores the estimated sigmas without spatial modelling


    float         * backgroundAppearanceSigmas;
    int           * probSpatialSum;
    int           * numberOfFGSamples;
    float         * pdfFG;
    float         * pdfBG;
    unsigned char * morphologicalFG;
    unsigned char * fgMasks;

    ColouredPixel<T, 3> * foregroundReferencePixels;

    float         * fgPriorProbability;

    ProcessingDeviceData ();


    virtual ~ProcessingDeviceData ();

    void allocateDeviceMemory (unsigned int width,
                               unsigned int height,
                               unsigned int channels,
                               unsigned int numImages,
                               unsigned int numForegroundReferenceFrames);

    void releaseDeviceMemory();
private:
    ProcessingDeviceData (const ProcessingDeviceData &other) { (void) other; }
};

struct ProcessingHostData {
    unsigned int      availableReference;
    std::ptrdiff_t    readIndex;
    std::ptrdiff_t    writeIndex;
    Thumbnail         thumbnail;
    ParticleFilter    particleFilter;
    dim3              gridSize;
    dim3              blockSize;
    cudaStream_t      stream;

    ProcessingHostData () {}

private:
    ProcessingHostData (const ProcessingHostData & other);
};



#endif // PROCESSINGDATATYPES_H
