#ifndef STAGE_H
#define STAGE_H

#include <QtCore/QThread>
#include <cstddef>
#include "imagedimensions.h"
class cudaArray;

class Stage : public QThread
{
public:

    // Returns the layer of the cudaArray at which the previous stage should begin
    // to write. Thus, if this function returns l and layers=n, the previous stage
    // must write in layers l, l+1, ..., l+n-2, l+n-1
    virtual std::ptrdiff_t startWrite () = 0;

    // Through this function, the previous stage notifies this stage that it has
    // completely written all the layers {written, written+1, ..., written+n-2, written+n-1}
    virtual void endWrite (std::ptrdiff_t written) = 0;
    virtual void notifyEndOfStream () = 0;


    // This tells the next stage to allocate the buffers it deems necessary to do its business with
    // images of the specified dimensions and number of channels
    virtual cudaArray* initializeBuffer (unsigned int width, unsigned int height, unsigned int layers) = 0;
    virtual cudaArray* initializeBuffer (ImageDimensions const &dim) = 0;
protected:
    Stage *next;
};

#endif // STAGE_H
